#[macro_use]
extern crate log;
extern crate rag;
extern crate serde;
#[macro_use]
extern crate serde_derive;
extern crate toml;

use std::error::Error;
use std::fs::File;
use std::io::Read;
use std::path::PathBuf;
use std::process::{Child, Command, ExitStatus, Stdio};

type SupResult<T> = Result<T, Box<dyn Error>>;

#[derive(Clone, Debug, Deserialize)]
struct App {
	name: String,
	command: String,
	args: Vec<String>,
	restart_on_success: Option<bool>,
	restart_on_failure: Option<bool>,
	restart_on_terminate: Option<bool>,
	wait: Option<bool>,
	hold: Option<bool>,
}

impl App {
	fn wait_start(&self) -> SupResult<()> {
		info!("starting application {}", &self.name);

		let mut command = Command::new(&self.command);
		command.args(&self.args);

		match command.output() {
			Ok(output) => {
				if !output.stdout.is_empty() {
					let s = String::from_utf8(output.stdout)?;
					for line in s.lines() {
						info!("[{}] stdout: {}", &self.name, line);
					}
				}
				if !output.stderr.is_empty() {
					let s = String::from_utf8(output.stderr)?;
					for line in s.lines() {
						info!("[{}] stderr: {}", &self.name, line);
					}
				}
				Ok(())
			}
			Err(e) => Err(Box::from(e)),
		}
	}
}

#[derive(Debug, Deserialize)]
struct Config {
	app: Vec<App>,
}

impl Config {
	fn load<T: Into<PathBuf>>(p: T) -> SupResult<Config> {
		let mut file = File::open(p.into())?;
		let mut buf = String::new();
		file.read_to_string(&mut buf)?;
		Ok(toml::from_str(&buf)?)
	}

	fn apps(&self) -> Vec<App> {
		self.app.clone()
	}
}

struct Proc {
	app: App,
	child: Child,
}

impl Proc {
	fn start(app: App) -> SupResult<Proc> {
		info!("starting application {}", &app.name);

		let mut command = Command::new(&app.command);
		command.stdout(Stdio::piped());
		command.stderr(Stdio::piped());
		command.args(&app.args);

		let child = command.spawn()?;

		Ok(Proc { app, child })
	}

	fn check_stdout(&mut self) -> Option<String> {
		let child_stdout = self.child.stdout.as_mut();
		if let Some(stdout) = child_stdout {
			let mut buf = String::new();
			match stdout.read_to_string(&mut buf) {
				Ok(size) => {
					if size > 0 {
						return Some(buf);
					} else {
						return None;
					}
				}
				Err(e) => {
					error!("couldn't read stdout from {}: {:?}", &self.app.name, e);
					return None;
				}
			}
		}
		None
	}

	fn check_stderr(&mut self) -> Option<String> {
		let child_stderr = self.child.stderr.as_mut();
		if let Some(stderr) = child_stderr {
			let mut buf = String::new();
			match stderr.read_to_string(&mut buf) {
				Ok(size) => {
					if size > 0 {
						return Some(buf);
					} else {
						return None;
					}
				}
				Err(e) => {
					error!("couldn't read stderr from {}: {:?}", &self.app.name, e);
					return None;
				}
			}
		}
		None
	}
}

fn main() {
	if let Err(e) = rag::init() {
		println!("FATAL: Error initializing logger: {:?}", e);
		std::process::exit(1);
	}

	if let Err(e) = run() {
		error!("FATAL: {:?}", e);
		std::process::exit(1);
	}
}

#[allow(clippy::cognitive_complexity)]
fn run() -> SupResult<()> {
	let config = Config::load("sup.toml")?;
	info!("loaded config: sup.toml");

	let mut procs: Vec<Proc> = Vec::new();
	let mut holds: Option<i8> = None;

	for app in config.apps() {
		if app.wait.unwrap_or(false) {
			if let Err(e) = app.wait_start() {
				error!("app failed to start: {}: {:?}", &app.name, e);
			}
		} else {
			let name = app.name.clone();
			match Proc::start(app) {
				Ok(proc) => {
					if proc.app.hold.unwrap_or(false) {
						holds = Some(holds.unwrap_or(0) + 1);
					}
					procs.push(proc);
				}
				Err(e) => error!("application failed to start: {}: {:?}", name, e),
			};
		}
	}

	if holds.unwrap_or(0) < 1 {
		error!("no holds configured");
		return Ok(());
	}

	while holds > Some(0) {
		let mut newprocs: Vec<Proc> = Vec::new();

		while let Some(mut proc) = procs.pop() {
			match proc.child.try_wait() {
				Ok(Some(status)) => {
					let hold = proc.app.hold.unwrap_or(false);
					if handle_exit(&proc.app, status) {
						let name = proc.app.name.clone();
						match Proc::start(proc.app) {
							Ok(p) => newprocs.push(p),
							Err(e) => {
								error!("error restarting {}: {:?}", name, e);
								if hold {
									holds = Some(holds.unwrap_or(0) - 1);
								}
							}
						};
					} else if hold {
						holds = Some(holds.unwrap_or(0) - 1);
					}
				}
				Ok(None) => {
					if let Some(s) = &proc.check_stdout() {
						for line in s.lines() {
							info!("[{}] stdout: {}", &proc.app.name, line);
						}
					}
					if let Some(s) = &proc.check_stderr() {
						for line in s.lines() {
							info!("[{}] stderr: {}", &proc.app.name, line);
						}
					}
					newprocs.push(proc);
				}
				Err(e) => error!("error: {:?}", e),
			}
		}
		procs = newprocs;
	}

	info!("shutting down");

	for mut proc in procs {
		let _ = proc.child.kill();
	}

	Ok(())
}

fn handle_exit(app: &App, status: ExitStatus) -> bool {
	if status.success() {
		info!("application exited: {}", &app.name);
		app.restart_on_success.unwrap_or(false)
	} else if let Some(code) = status.code() {
		info!("application failed: {} ({})", &app.name, code);
		app.restart_on_failure.unwrap_or(false)
	} else {
		info!("application terminated: {}", &app.name);
		app.restart_on_terminate.unwrap_or(false)
	}
}
